/*
 * Copyright 2014-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
/*
 * TERMS OF USE - EASING EQUATIONS
 * Open source under the BSD License.
 *
 * Copyright © 2001 Robert Penner
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 * Neither the name of the author nor the names of contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package ej.motion.bounce;

import ej.motion.Function;

/**
 * Represents a bounce motion: starts at a steady speed, then slows down to the stop instruction then go backward with
 * the same effect with less initial speed then go forward and so on until it stops.
 */
public class BounceEaseOutFunction implements Function {

	/**
	 * Singleton to avoid creating several instances.
	 */
	public static final BounceEaseOutFunction INSTANCE = new BounceEaseOutFunction();

	private static final float N1 = 7.5625f;
	private static final float D1 = 2.75f;
	private static final float FIRST_LIMIT = 1.0f;
	private static final float SECOND_LIMIT = 2.0f;
	private static final float THIRD_LIMIT = 2.5f;
	private static final float SECOND_FACTOR = 1.5f;
	private static final float THIRD_FACTOR = 2.25f;
	private static final float FOURTH_FACTOR = 2.625f;
	private static final float SECOND_INCREMENT = 0.75f;
	private static final float THIRD_INCREMENT = 0.9375f;
	private static final float FOURTH_INCREMENT = 0.984375f;

	private BounceEaseOutFunction() {
	}

	@Override
	public float computeValue(float t) {
		float increment;
		if (t < FIRST_LIMIT / D1) {
			increment = 0.0f;
		} else if (t < SECOND_LIMIT / D1) {
			t -= SECOND_FACTOR / D1;
			increment = SECOND_INCREMENT;
		} else if (t < THIRD_LIMIT / D1) {
			t -= THIRD_FACTOR / D1;
			increment = THIRD_INCREMENT;
		} else {
			t -= FOURTH_FACTOR / D1;
			increment = FOURTH_INCREMENT;
		}
		return N1 * t * t + increment;
	}

}
